#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>

#include "socket.h"

static void error_exit(char *message)
{
  fprintf(stderr, message);
  exit(1);
}

int socket_open_server(int port)
{
  int s;
  struct sockaddr_in addr;
  int on = 1;

  s = socket(AF_INET, SOCK_STREAM, 0);
  if (s < 0)
    error_exit("socket() failed.\n");

  if (setsockopt(s, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on)) < 0)
    error_exit("setsockopt() failed.\n");

  memset(&addr, 0, sizeof(addr));
  addr.sin_family = AF_INET;
  addr.sin_port = htons(port);
  addr.sin_addr.s_addr = htonl(INADDR_ANY);

  if (bind(s, (struct sockaddr *)&addr, sizeof(addr)) < 0)
    error_exit("bind() failed.\n");

  if (listen(s, 5) < 0)
    error_exit("listen() failed.\n");

  return s;
}

int socket_accept_server(int sockt)
{
  int s;
  struct sockaddr_in addr;
  socklen_t addrlen;
  fd_set fds;
  struct timeval t;

  FD_ZERO(&fds);
  FD_SET(sockt, &fds);
  t.tv_sec  = 0;
  t.tv_usec = 0;
  if (select(sockt + 1, &fds, NULL, NULL, &t) <= 0)
    return -1;
  if (!FD_ISSET(sockt, &fds))
    return -1;

  memset(&addr, 0, sizeof(addr));
  addrlen = sizeof(addr);
  s = accept(sockt, (struct sockaddr *)&addr, &addrlen);
  if (s < 0)
    error_exit("accept() failed.\n");

  return s;
}

int socket_open_dgram_server(int port)
{
  int s;
  struct sockaddr_in addr;
  int on = 1;

  s = socket(AF_INET, SOCK_DGRAM, 0);
  if (s < 0)
    error_exit("socket() failed.\n");

  if (setsockopt(s, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on)) < 0)
    error_exit("setsockopt() failed.\n");

  memset(&addr, 0, sizeof(addr));
  addr.sin_family = AF_INET;
  addr.sin_port = htons(port);
  addr.sin_addr.s_addr = htonl(INADDR_ANY);

  if (bind(s, (struct sockaddr *)&addr, sizeof(addr)) < 0)
    error_exit("bind() failed.\n");

  return s;
}

int socket_open_client(char *hostname, int port)
{
  int s;
  struct hostent *host;
  struct sockaddr_in addr;

  host = gethostbyname(hostname);
  if (host == NULL)
    error_exit("gethostbyname() failed.\n");

  s = socket(AF_INET, SOCK_STREAM, 0);
  if (s < 0)
    error_exit("socket() failed.\n");

  memset(&addr, 0, sizeof(addr));
  addr.sin_family = AF_INET;
  addr.sin_port = htons(port);
  memcpy(&addr.sin_addr, host->h_addr, host->h_length);

  if (connect(s, (struct sockaddr *)&addr, sizeof(addr)) < 0)
    error_exit("connect() failed.\n");

  return s;
}

int socket_read(int s, char *buffer, int size, int wait_usec)
{
  fd_set fds;
  struct timeval t;
  int r;

  if (wait_usec >= 0) {
    FD_ZERO(&fds);
    FD_SET(s, &fds);
    t.tv_sec  = wait_usec / 1000000;
    t.tv_usec = wait_usec % 1000000;
    r = select(s + 1, &fds, NULL, NULL, &t);
    if (r <= 0)
      return 0;
    if (!FD_ISSET(s, &fds))
      return 0;
  }

  r = read(s, buffer, size);
  if (r < 0)
    return 0;
  return (r == 0) ? -1 : r;
}

int socket_write(int s, char *buffer, int size)
{
  return write(s, buffer, size);
}

static char *skip_space(char *p)
{
  while (*p) {
    if (!isspace(*p))
      break;
    p++;
  }
  return p;
}

#if 0
static char *skip_nospace(char *p)
{
  while (*p) {
    if (isspace(*p))
      break;
    p++;
  }
  return p;
}
#endif

static char *skip_noquote(char *p, int quote)
{
  int escape = 0;
  char *q;

  q = p;
  while (*p) {
    if (!escape) {
      if (*p == '\\') {
	escape = 1;
	*(p++) = '\0';
	continue;
      }
      if (quote) {
	if (*p == quote) break;
      } else {
	if (isspace(*p)) break;
      }
    }
    if ((*p == '\n') || (*p == '\r'))
      break;
    escape = 0;

    if (p != q) *q = *p;
    p++;
    q++;
    if (p != q) *q = '\0';
  }

  return p;
}

char **socket_readline(int *s, int wait_usec)
{
  static int buffer_offset = 0;
  static int buffer_size = 0;
  static char buffer[1024];
  static char *argv[16];
  int r, i, quote;
  char *p;

  if (buffer_offset > 0) {
    for (i = 0; i < buffer_size; i++) {
      buffer[i] = buffer[buffer_offset + i];
    }
    buffer_offset = 0;
  }

  r = socket_read(*s, buffer + buffer_size,
		  sizeof(buffer) - buffer_size, wait_usec);
  if (r < 0) {
    close(*s);
    *s = -1;
    return NULL;
  }

  buffer_size += r;

  for (i = 0; i < buffer_size; i++) {
    if ((buffer[i] == '\n') || (buffer[i] == '\r'))
      break;
  }
  if (i == buffer_size) {
    if (buffer_size == sizeof(buffer))
      buffer_size = 0;
    return NULL;
  }
  buffer[i] = '\0';

  buffer_offset = i + 1;
  buffer_size -= buffer_offset;

  p = buffer;
  i = 0;

  while (1) {
    p = skip_space(p);
    if (*p == '\0')
      break;

    quote = 0;
    if ((*p == '\'') || (*p == '\"')) {
      quote = *p;
      p++;
    }

    argv[i++] = p;
    if (i == sizeof(argv) / sizeof(argv[0]))
      return NULL;

    p = skip_noquote(p, quote);
    if (*p == '\0')
      break;

    *(p++) = '\0';
  }

  argv[i] = NULL;

  return argv;
}
